# Copyright (C) 2009 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from usbcreator.backends.base import Backend
from usbcreator.misc import *
import os
import ctypes
import ctypes.wintypes
import logging

DRIVE_TYPES = ['unknown',
               'no_root_dir',
               'removeable',
               'localdisk',
               'remotedisk',
               'cdrom',
               'ramdisk']

class WindowsBackend(Backend):
    def __init__(self):
        Backend.__init__(self)

    # Device processing.

    def detect_devices(self):
        logical_drives = ctypes.windll.kernel32.GetLogicalDrives()

        for i in range(26):
            if (logical_drives >> i) & 0x01:
                drive = u'%s:\\' % chr(i + 65)
                self._device_added(drive)

    def _device_added(self, drive):
        drive_buf = ctypes.create_unicode_buffer(512)
        fs_buf = ctypes.create_unicode_buffer(512)
        vol_i = ctypes.windll.kernel32.GetVolumeInformationW
        drive_type = ctypes.windll.kernel32.GetDriveTypeW(drive)
        if vol_i(drive, drive_buf, 512, None, None, None, fs_buf, 512):
            volume = drive_buf.value
            fs = fs_buf.value
        else:
            logging.error('Could not get volume information for %s' % drive)
            return

        if drive_type > len(DRIVE_TYPES):
               drive_type = 0
        drive_type = DRIVE_TYPES[drive_type]

        if drive_type == 'cdrom':
            if os.path.exists(os.path.join(drive, '.disk/info')):
                self.sources[drive] = {
                    'device' : drive,
                    'size' : fs_size(drive)[0],
                    'label' : volume,
                    'type' : SOURCE_CD,
                }
                if callable(self.source_added_cb):
                    self.source_added_cb(drive)
        elif drive_type == 'removeable': # and fs == u'fat32':
            tot, free = fs_size(drive)
            self.targets[drive] = {
                'capacity' : tot,
                'free' : free,
                'device' : drive,
                # FIXME evand 2009-07-16: How does Windows get the volume name?
                # GetVolumeInformation returns NULL here.
                'label' : '',
                'status' : CAN_USE,
            }
            if callable(self.target_added_cb):
                self.target_added_cb(drive)
        else:
            tot, free = fs_size(drive)
            logging.debug('Not adding %s (%s %s %s volume, %d total bytes, '
                          '%d bytes free)' %
                          (drive, volume, drive_type, fs, tot, free))

    def add_image(self, filename):
        logging.debug('Backend told to add: %s' % filename)
        if filename in self.sources:
            logging.warn('Source already added.')
            # TODO evand 2009-07-27: Scroll to source and select.
            return

        extension = os.path.splitext(filename)[1]
        # TODO evand 2009-07-25: What's the equivalent of `file` on Windows?
        # Going by extension is a bit rough.
        if not extension:
            logging.error('File did not have an extension.  '
                          'Could not determine file type.')
            # TODO evand 2009-07-26: Error dialog.
            return

        extension = extension.lower()
        if extension == '.iso':
            # XXX evand 2009-07-25: Using 7zip until cdkit gets mingw support.
            cmd = ['7z', 'x', filename, '.disk/info', '-so']
            try:
                output = popen(cmd, stderr=None)
                if output:
                    self.sources[filename] = {
                        'device' : filename,
                        'size' : os.path.getsize(filename),
                        'label' : output,
                        'type' : SOURCE_ISO,
                    }
                    if callable(self.source_added_cb):
                        self.source_added_cb(filename)
            except USBCreatorProcessException:
                logging.error('Could not extract .disk/info.')
                # TODO evand 2009-07-26: Error dialog.
                return
        elif extension == '.img':
            self.sources[filename] = {
                'device' : filename,
                'size' : os.path.getsize(filename),
                'label' : '',
                'type' : SOURCE_IMG,
            }
            if callable(self.source_added_cb):
                self.source_added_cb(filename)
        else:
            logging.error('Filename extension type not supported.')

    def format(self, drive):
        assert drive == 'E:\\'
        assert drive in self.targets
        # TODO evand 2009-07-16: Windows does not provide an API for
        # formatting.  It offers a method to bring up the format dialog, but
        # that will not work as there is no way to force it to the FAT32
        # filesystem.
        # One can format a volume using WMI, but that would pull in a
        # dependency on Mark Hammond's pywin32.

