#!/usr/bin/python

# Copyright (C) 2009 Roderick B. Greening <roderick.greening@gmail.com>
# 
# Based in part on work by:
#  David Edmundson <kde@davidedmundson.co.uk>
#  Canonical Ltd. USB Creator Team
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os

from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4 import uic
from PyKDE4.kdeui import KIcon, KMessageBox
from PyKDE4.kdecore import KProcess, KStandardDirs, KUrl
from PyKDE4.kio import KFileDialog
from PyKDE4.solid import Solid

from usbcreator.frontends.kde.translate import translate
uic.properties.Properties._string = translate
import gettext
import locale
import logging

from usbcreator.frontends.base import Frontend
from usbcreator.misc import *
from Queue import Queue

queue = Queue()
def thread_wrap(func):
    '''Decorator for functions that will be called by another thread.'''
    def wrap(*args):
        queue.put(lambda: func(*args))
    return wrap

class KdeFrontend(Frontend, QObject):
    @classmethod
    def startup_failure(cls, message):
        KMessageBox.sorry(None, message, "", KMessageBox.Notify)

    @classmethod
    def DBusMainLoop(cls):
        from dbus.mainloop.qt import DBusQtMainLoop
        DBusQtMainLoop(set_as_default=True)

    def __init__(self, backend, img=None, persistent=True):
        QObject.__init__(self)

        #our passed vars - keep them private
        self.__persistent = persistent
        self.__img = img

        # Perform some initialization
        self.__initPrivateVars()
        self.__initUI()

        # FIXME: preload the source/taget treeview...
        #self.setup_sources_treeview()
        #self.setup_targets_treeview()

        #enable the backend
        self.__backend = backend

        # Connect to backend signals.
        self.__backend.source_added_cb = self.add_source
        self.__backend.target_added_cb = self.add_target
        self.__backend.source_removed_cb = self.remove_source
        self.__backend.target_removed_cb = self.remove_target
        self.__backend.failure_cb = self.failure
        self.__backend.success_cb = self.success
        self.__backend.install_progress_cb = self.progress
        self.__backend.install_progress_message_cb = self.progress_message
        self.__backend.retry_cb = self.retry

        #add any file sources passed
        if self.__img is not None:
            self.__backend.add_image(self.__img)

        def test_func(*a):
            if not queue.empty():
                func = queue.get_nowait()
                func()
                queue.task_done()
            return True
        self.queue_processor = self.add_timeout(500, test_func, None)
        
        backend.detect_devices()

    def __initPrivateVars(self):
        """Initialize Private Variables"""

        # main window
        self.__mainWindow = QDialog()

        # ui file
        self.__mainWindow_ui = "usbcreator-kde.ui"

        # init Backend to None - easier to debug...
        self.__backend = None

    def __initUI(self):
        """Initialize the interface"""

        # Locate the ui for the main window and load it.
        if 'USBCREATOR_LOCAL' in os.environ:
            appdir = os.path.join(os.getcwd(), 'gui')
        else:
            file =  KStandardDirs.locate("appdata", self.__mainWindow_ui)
            appdir = file.left(file.lastIndexOf("/"))
        uic.loadUi(appdir + "/" + self.__mainWindow_ui, self.__mainWindow)

        #set default persist size
        self.__mainWindow.ui_persist_label.setText(format_mb_size(128))

        #set persistent ui elements state
        if self.__persistent:
            self.__mainWindow.ui_persist_enabled.setChecked(True)
            self.__mainWindow.ui_persist_text.setEnabled(True)
            self.__mainWindow.ui_persist_slider.setEnabled(True)
            self.__mainWindow.ui_persist_label.setEnabled(True)
        else:
            self.__mainWindow.ui_persist_disabled.setChecked(True)
            self.__mainWindow.ui_persist_text.setDisabled(True)
            self.__mainWindow.ui_persist_slider.setDisabled(True)
            self.__mainWindow.ui_persist_label.setDisabled(True)

        #disable the start button and persist frame by default
        self.__mainWindow.ui_persist_frame.setEnabled(False)
        self.__mainWindow.ui_start_button.setEnabled(False)


        #add some buttons
        self.__mainWindow.ui_quit_button.setIcon(KIcon("application-exit"))
        self.__mainWindow.ui_start_button.setIcon(KIcon("dialog-ok-apply"))
        self.__mainWindow.ui_add_source.setIcon(KIcon("media-optical"))
        self.__mainWindow.ui_format_dest.setIcon(KIcon("drive-removable-media-usb-pendrive"))

        #set up signals
        self.connect(self.__mainWindow.ui_add_source,SIGNAL('clicked()'),
            self.add_file_source_dialog)
        self.connect(self.__mainWindow.ui_persist_slider,SIGNAL('valueChanged(int)'),
            lambda value: self.__mainWindow.ui_persist_label.setText(format_mb_size(value)))
        self.connect(self.__mainWindow.ui_quit_button,SIGNAL('clicked()'),
            self.quit)
        self.connect(self.__mainWindow.ui_start_button,SIGNAL('clicked()'),
            self.install)
        self.connect(self.__mainWindow.ui_dest_list,SIGNAL(
            'currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)'),
            self.dest_selection_changed)
        self.connect(self.__mainWindow.ui_source_list,SIGNAL(
            'currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)'),
            self.source_selection_changed)
        self.connect(self.__mainWindow.ui_format_dest,SIGNAL('clicked()'),
            self.format_dest_clicked)

        self.progress_bar = QProgressDialog("","Cancel",0,100,self.__mainWindow)
        #prevent progress bar from emitting reset on reaching max value (and auto closing)
        self.progress_bar.setAutoReset(False)
        #force immediate showing, rather than waiting...
        self.progress_bar.setMinimumDuration(0)
        #must disconnect the canceled() SIGNAL, otherwise the progress bar is actually destroyed
        self.disconnect(self.progress_bar,SIGNAL('canceled()'),self.progress_bar.cancel)
        #now we connect our own signal to display a warning dialog instead
        self.connect(self.progress_bar,SIGNAL('canceled()'),self.warning_dialog)

        #show the window
        self.__mainWindow.show()

    def __timeout_callback(self, func, *args):
        '''Private callback wrapper used by add_timeout'''

        timer = self.sender()
        active = func(*args)
        if not active:
            timer.stop()

    def add_timeout(self, interval, func, *args):
        '''Add a new timer for function 'func' with optional arguments. Mirrors a 
        similar gobject call timeout_add.'''

        # FIXME: now that we are part of a Qt object, we may be able to alter for builtin timers
        timer = QTimer()
        QObject.connect(timer,
            SIGNAL('timeout()'),
            lambda: self.__timeout_callback(func, *args))
        timer.start(interval)

        return timer

    def delete_timeout(self, timer):
        '''Remove the specified timer'''

        if timer.isActive():
            return False
        timer.stop()
        return True

    def add_target(self, target):
        logging.debug('add_target: %s' % str(target))
        new_item = QTreeWidgetItem(self.__mainWindow.ui_dest_list)
        new_item.setData(0,Qt.UserRole,QVariant(target))
        # FIXME:
        # the new_item lines should be auto triggered onChange to the
        # TreeWidget when new_item is appended. aka setup_targets_treeview +
        # callback
        new_item.setText(0,target)
        new_item.setIcon(0,KIcon("drive-removable-media-usb-pendrive"))

        item = self.__mainWindow.ui_dest_list.currentItem()
        if not item:
            item = self.__mainWindow.ui_dest_list.topLevelItem(0)
            if item:
                self.__mainWindow.ui_dest_list.setCurrentItem(item,True)

        # populate from device data
        if self.__backend is not None:
            dev = self.__backend.targets[target]
            new_item.setText(0,dev['device'])
            new_item.setText(1,dev['label'])
            new_item.setText(2,format_size(dev['capacity']))
            new_item.setText(3,format_size(dev['free']))
 
    def remove_target(self, target):
        for i in range(0,self.__mainWindow.ui_dest_list.topLevelItemCount()):
            item = self.__mainWindow.ui_dest_list.topLevelItem(i)
            if item.data(0,Qt.UserRole).toString() == target:
                self.__mainWindow.ui_dest_list.removeItemWidget(item,0)
                break

        if not self.__mainWindow.ui_dest_list.currentItem():
            item = self.__mainWindow.ui_dest_list.topLevelItem(0)
            if item:
                self.__mainWindow.ui_dest_list.setCurrentItem(item,True)

    def add_source(self, source):
        logging.debug('add_source: %s' % str(source))
        new_item = QTreeWidgetItem(self.__mainWindow.ui_source_list)
        new_item.setData(0,Qt.UserRole,QVariant(source))
        # FIXME:
        # the new_item lines should be auto triggered onChange to the TreeWidget
        # when new_item is appended. aka setup_targets_treeview + callback
        new_item.setText(0,source)
        new_item.setIcon(0,KIcon("media-optical"))

        item = self.__mainWindow.ui_source_list.currentItem()
        if not item:
            item = self.__mainWindow.ui_source_list.topLevelItem(0)
            if item:
                self.__mainWindow.ui_source_list.setCurrentItem(item,True)

        # how does this all get added? here or elsewhere...
        # populate from device data
        if self.__backend is not None:
            new_item.setText(0,self.__backend.sources[source]['device'])
            new_item.setText(1,self.__backend.sources[source]['label'])
            new_item.setText(2,format_size(self.__backend.sources[source]['size']))
 
    def remove_source(self, source):
        for i in range(0,self.__mainWindow.ui_source_list.topLevelItemCount()):
            item = self.__mainWindow.ui_source_list.topLevelItem(i)
            if item.data(0,Qt.UserRole).toString() == source:
                self.__mainWindow.ui_source_list.removeItemWidget(item,0)
                break

        if not self.__mainWindow.ui_source_list.currentItem():
            item = self.__mainWindow.ui_source_list.topLevelItem(0)
            if item:
                self.__mainWindow.ui_source_list.setCurrentItem(item,True)

    def get_source(self):
        '''Returns the UDI of the selected source image.'''
        item = self.__mainWindow.ui_source_list.currentItem()
        if item:
            # Must deal in str and not QString for backend
            source = str(item.data(0,Qt.UserRole).toString())
            return source
        else:
            logging.debug('No source selected.')
            return ''

    def get_target(self):
        '''Returns the UDI of the selected target disk or partition.'''
        item = self.__mainWindow.ui_dest_list.currentItem()
        if item:
            # Must deal in str and not QString for backend
            dest = str(item.data(0,Qt.UserRole).toString())
            return dest
        else:
            logging.debug('No target selected.')
            return ''

    def get_persistence(self):
        if (self.__mainWindow.ui_persist_enabled.isChecked() and
            self.__mainWindow.ui_persist_frame.isEnabled()):
            val = self.__mainWindow.ui_persist_slider.value()
            return int(val)
        else:
            return 0

    def get_solid_drive(self, udi):
        deviceList = Solid.Device.allDevices()
        drive = {}
        for device in deviceList:
            if device.isDeviceInterface(Solid.DeviceInterface.StorageVolume):
                volume = device.asDeviceInterface (Solid.DeviceInterface.StorageVolume)
                if str(device.udi()) == udi:
                    drive['icon'] = device.icon()
                    drive['display_name'] = volume.label()
        return drive

    def setup_sources_treeview(self):
        def column_data_func(item, column):
            if not self.__backend:
                return
            udi = str(item.data(0,Qt.UserRole).toString())
            dev = self.__backend.sources[udi]
            if column == 0:
                drive = self.get_solid_drive(udi)
                if drive:
                    cell.set_property('text', drive['display_name'])
                else:
                    cell.set_property('text', dev['device'])
            elif column == 1:
                cell.set_property('text', dev['label'])
            elif column == 2:
                cell.set_property('text', format_size(dev['capacity']))

        def pixbuf_data_func(column, cell, model, iterator):
            if not self.__backend:
                return
            udi = model[iterator][0]
            drive = self.get_gnome_drive(udi)
            type = self.__backend.sources[udi]['type']
            if drive:
                cell.set_property('icon-name', drive['icon'])
            elif type == Backend.ISO:
                cell.set_property('stock-id', gtk.STOCK_CDROM)
            elif type == Backend.IMG:
                cell.set_property('stock-id', gtk.STOCK_HARDDISK)
            else:
                cell.set_property('stock-id', None)

        list_store = gtk.ListStore(str)
        self.source_treeview.set_model(list_store)

        cell_name = gtk.CellRendererText()
        cell_pixbuf = gtk.CellRendererPixbuf()
        column_name = gtk.TreeViewColumn(_('CD-Drive/Image'))
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        column_name.pack_start(cell_pixbuf, expand=False)
        column_name.pack_start(cell_name, expand=True)
        self.source_treeview.append_column(column_name)
        column_name.set_cell_data_func(cell_name, column_data_func, 0)
        column_name.set_cell_data_func(cell_pixbuf, pixbuf_data_func)

        cell_version = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn(_('OS Version'), cell_version)
        column_name.set_cell_data_func(cell_version, column_data_func, 1)
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.source_treeview.append_column(column_name)

        cell_size = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn(_('Size'), cell_size)
        column_name.set_cell_data_func(cell_size, column_data_func, 2)
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.source_treeview.append_column(column_name)

        selection = self.source_treeview.get_selection()
        selection.connect('changed', lambda x: self.__backend.refresh_targets())

        # Drag and drop support.
        # FIXME evand 2009-04-28: Anything can be dropped on the source
        # treeview.  Ideally, the user should only be able to drop ISO and IMG
        # files.

        #def motion_cb(wid, context, x, y, time):
        #    context.drag_status(gtk.gdk.ACTION_COPY, time)
        #    return True

        #def drop_cb(w, context, x, y, time):
        #    target_list = w.drag_dest_get_target_list()
        #    target = w.drag_dest_find_target(context, target_list)
        #    selection = w.drag_get_data(context, target)
        #    context.finish(True, True)
        #    return True

        #def data_received_cb(w, context, x, y, selection, target_type, timestamp):
        #    # FIXME evand 2009-04-28: Use the GNOME VFS?  Test with a sshfs
        #    # nautilus window.
        #    file = selection.data.strip('\r\n\x00')
        #    if file.startswith('file://'):
        #        file = file[7:]
        #    elif file.startswith('file:'):
        #        file = file[5:]
        #    self.__backend.add_file_source(file)

        #self.source_treeview.drag_dest_set(gtk.gdk.ACTION_DEFAULT,
        #    [('text/uri-list', 0, 600)], gtk.gdk.ACTION_COPY)
        #self.source_treeview.connect('drag_motion', motion_cb)
        #self.source_treeview.connect('drag_drop', drop_cb)
        #self.source_treeview.connect('drag-data-received', data_received_cb)

    def update_target(self, udi):
        for i in range(0,self.__mainWindow.ui_dest_list.topLevelItemCount()):
            item = self.__mainWindow.ui_dest_list.topLevelItem(i)
            if str(item.data(0,Qt.UserRole).toString()) == udi:
                self.__mainWindow.ui_dest_list.emit(
                    SIGNAL('itemChanged(item,0)'))
                break

    def source_selection_changed(self, current_item, prev_item):
        '''The selected image has changed we need to refresh targets'''
        #if self.__backend:
        #    self.__backend.refresh_targets()
        pass

    def dest_selection_changed(self, current_item, prev_item):
        '''The selected partition has changed and the bounds on the persistence
        slider need to be changed, or the slider needs to be disabled, to
        reflect the amount of free space on the partition.'''

        # Disable some elements
        self.__mainWindow.ui_persist_frame.setEnabled(False)
        self.__mainWindow.ui_start_button.setEnabled(False)

        if not self.__backend:
            return

        if current_item is None:
            return

        # Must deal in str and not QString for backend
        target_udi = str(current_item.data(0,Qt.UserRole).toString())
        source_udi = self.get_source()

        if not source_udi:
            return

        #enable start button
        self.__mainWindow.ui_start_button.setEnabled(True)

        source = self.__backend.sources[source_udi]
        # We need to calculate how much, if any, extra space can be used for
        # the persistence file.
        target = self.__backend.targets[target_udi]
        persist_max = (target['free'] - source['size']) / 1024 / 1024
        if persist_max > MIN_PERSISTENCE:
            self.__mainWindow.ui_persist_frame.setEnabled(True)
            self.__mainWindow.ui_persist_slider.setRange(MIN_PERSISTENCE, persist_max)

    def setup_targets_treeview(self):
        def column_data_func(item, column):
            if not self.__backend:
                return
            udi = str(item.data(0,Qt.UserRole).toString())
            dev = self.__backend.targets[udi]
            drive = self.get_solid_drive(udi)
            if column == 0:
                if drive:
                    cell.set_property('text', drive['display_name'])
                else:
                    cell.set_property('text', dev['device'])
            elif column == 1:
                cell.set_property('text', dev['label'])
            elif column == 2:
                cell.set_property('text', format_size(dev['capacity']))
            elif column == 3:
                cell.set_property('text', format_size(dev['free']))

        def pixbuf_data_func(column, cell, model, iterator):
            if not self.__backend:
                return
            udi = model[iterator][0]
            status = self.__backend.targets[udi]['status']

            if status == Backend.NEED_SPACE:
                cell.set_property('stock-id', gtk.STOCK_DIALOG_WARNING)
            elif status == Backend.CANNOT_USE:
                # TODO evand 2009-05-05: Implement disabled rows as a
                # replacement?
                cell.set_property('stock-id', gtk.STOCK_DIALOG_ERROR)
            else:
                drive = self.get_gnome_drive(udi)
                if drive:
                    cell.set_property('icon-name', drive['icon'])
                else:
                    cell.set_property('stock-id', None)

        list_store = gtk.ListStore(str, int)
        self.dest_treeview.set_model(list_store)

        column_name = gtk.TreeViewColumn()
        column_name.set_title(_('Device'))
        cell_name = gtk.CellRendererText()
        cell_pixbuf = gtk.CellRendererPixbuf()
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        column_name.pack_start(cell_pixbuf, expand=False)
        column_name.pack_start(cell_name, expand=True)
        self.dest_treeview.append_column(column_name)
        column_name.set_cell_data_func(cell_name, column_data_func, 0)
        column_name.set_cell_data_func(cell_pixbuf, pixbuf_data_func)

        cell_name = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn(_('Label'), cell_name)
        column_name.set_cell_data_func(cell_name, column_data_func, 1)
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.dest_treeview.append_column(column_name)

        cell_capacity = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn(_('Capacity'), cell_capacity)
        column_name.set_cell_data_func(cell_capacity, column_data_func, 2)
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.dest_treeview.append_column(column_name)

        cell_free = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn(_('Free Space'), cell_free)
        column_name.set_cell_data_func(cell_free, column_data_func, 3)
        column_name.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.dest_treeview.append_column(column_name)

        selection = self.dest_treeview.get_selection()
        selection.connect('changed', self.dest_selection_changed)

    def add_file_source_dialog(self):
        filename = ''
        filter = QString('*.iso|' + _('CD Images') + '\n*.img|' + _('Disk Images'))
        # FIXME: should set the default path KUrl to users home dir...
        # This is all screwy as its run as root under kdesudo... Home = root and not user.. blarg!
        # Need to convert to plain string for backend to work
        filename = str(KFileDialog.getOpenFileName(KUrl(),filter))
        if not filename:
          return
        self.__backend.add_image(filename)

    def install(self):
        source = self.get_source()
        target = self.get_target()
        persist = self.get_persistence()
        if (source and target):
            self.__mainWindow.hide()
            starting_up = _('Starting up')
            self.progress_bar.setLabelText(starting_up)
            self.progress_bar.show()
            self.__backend.install(source, target, persist)
        else:
            message = _('You must select both source image and target device first.')
            self.notify(message)

    @thread_wrap
    def progress(self, complete, remaining, speed):
        # Updating value cause dialog to re-appear from hidden (dunno why)
        if not self.progress_bar.isHidden():
            self.progress_bar.setValue(int(complete))
    
    @thread_wrap
    def progress_message(self, message):
        self.progress_bar.setLabelText(message)

    def quit(self, *args):
        self.__backend.cancel_install()
        sys.exit(0)

    @thread_wrap
    def failure(self, message=None):
        logging.critical('Installation failed.')
        self.progress_bar.hide()
        if title:
            logging.critical('Install failed: ' + message)
        else:
            logging.critical('Install failed')
        KMessageBox.error(self.__mainWindow, message)
        sys.exit(1)

    @thread_wrap
    def success(self):
        '''Install completed'''
        self.progress_bar.hide()
        text = _('Installation is complete.  You may now reboot your '
                 'computer with this USB thumb drive inserted to boot '
                 'Ubuntu.')
        KMessageBox.information(self.__mainWindow, text)
        sys.exit(0)

    @thread_wrap
    def retry(self):
        # TODO evand 2009-08-02: Implement a retry dialog.
        raise NotImplementedError

    def notify(self,title):
        KMessageBox.sorry(self.__mainWindow,title)

    def warning_dialog(self):
        '''A warning dialog to show when progress dialog cancel is pressed'''

        caption = _('Quit the installation?')
        text = _('Do you really want to quit the installation now?')

        #hide the progress bar - install will still continue in bg
        self.progress_bar.hide()

        res = KMessageBox.warningYesNo(self.__mainWindow,text,caption)

        if res == KMessageBox.Yes:
            self.quit()

        #user chose not to quit, so re-show progress bar
        self.progress_bar.show()

    def format_dest_clicked(self):
        # FIXME evand 2009-04-30: This needs a big warning dialog.
        item = self.__mainWindow.ui_dest_list.currentItem()
        if not item:
            return
        udi = str(item.data(0,Qt.UserRole).toString())
        self.__backend.format(udi)

